local JSON = require "ejoysdk_lua.ejoysdk_json"
local CONFIG = require "ejoysdk_lua.ejoysdk_config"
local Class = require "ejoysdk_lua.ejoysdk_class"

local M
if _ejoysdk.os and _ejoysdk.os() == 'android' then
    M = require "ejoysdk_lua.ejoysdk_android"
    CONFIG.set_config('os', 'android')
else
    M = require "ejoysdk_lua.ejoysdk_ios"
    CONFIG.set_config('os', 'ios')
end

-- 老版本没这个函数报错了
if _ejoysdk.bin_version then
    CONFIG.set_config('bin_version', _ejoysdk.bin_version())
else
    CONFIG.set_config('bin_version', 1)
end

M.CONFIG = CONFIG

local HTTP = M.HTTP
local HttpDns = {}
M.HttpDns = HttpDns

local function split_string(str, sep_opt)
    local sep, fields = sep_opt or " ", {}
    local pattern = string.format("([^%s]+)", sep)
    string.gsub(str, pattern, function(c) fields[#fields+1] = c end)
    return fields
end

HTTP.Header = {
    New = function(headers)
        local obj = {}
        if headers then
            for k, v in pairs(headers) do
                obj[k:lower()] = v
            end
        end
        return setmetatable(obj, HTTP.Header)
    end,
    __index = function(self, key)
        return rawget(self, key:lower())
    end,
    __newindex = function(self, key, value)
        rawset(self, key:lower(), value)
    end
}

function HTTP.check_and_update_headers(headers)
    local ret = HTTP.Header.New(headers)

    ret['User-Agent'] = ret['User-Agent'] or M.PLATFORM.HTTP_UA
    ret['Accept-Charset'] = ret['Accept-Charset'] or "UTF-8"
    return ret
end

local escape = function(str)
    str = string.gsub(str, "\n", "\r\n")
    str = string.gsub(str, "([^A-Za-z0-9%_%.%-%~])", -- locale independent
        function (c) return string.format ("%%%02X", string.byte(c)) end)
    str = string.gsub(str, " ", "+")
    return str
end

local decode = function(str)
    str = str:gsub('+', ' ')
    return (str:gsub("%%(%x%x)", function(c)
        return string.char(tonumber(c, 16))
    end))
end

-- urlendocde 只支持1层嵌套数组
local urlencode = function(query)
    local ret = {}
    for k, v in pairs(query) do
        if type(v) ~= 'table' then
            v = {v}
        end

        for _, val in ipairs(v) do
            local val_type = type(val)
            assert(val_type ~= 'table')
            if val_type == 'number' or val_type == 'boolean' then
                val = tostring(val)
            end
            ret[#ret + 1] = escape(k) .. '=' .. escape(val)
        end
    end

    return table.concat(ret, '&')
end
HTTP.urlencode = urlencode

-- from https://github.com/golgote/neturl
function HTTP.parse_query(str, sep)
    sep = sep or '&'

	local values = {}
	for key_str, val in str:gmatch(string.format('([^%q=]+)(=*[^%q=]*)', sep, sep)) do
		local key = decode(key_str)
		local keys = {}
		key = key:gsub('%[([^%]]*)%]', function(v)
				-- extract keys between balanced brackets
				if string.find(v, "^-?%d+$") then
					v = tonumber(v)
				else
					v = decode(v)
				end
				table.insert(keys, v)
				return "="
		end)
		key = key:gsub('=+.*$', "")
		key = key:gsub('%s', "_") -- remove spaces in parameter name
		val = val:gsub('^=+', "")

		if not values[key] then
			values[key] = {}
		end
		if #keys > 0 and type(values[key]) ~= 'table' then
			values[key] = {}
		elseif #keys == 0 and type(values[key]) == 'table' then
			values[key] = decode(val)
		end

		local t = values[key]
		for i,k in ipairs(keys) do
			if type(t) ~= 'table' then
				t = {}
			end
			if k == "" then
				k = #t+1
			end
			if not t[k] then
				t[k] = {}
			end
			if i == #keys then
				t[k] = decode(val)
			end
			t = t[k]
		end
	end
	setmetatable(values, { __tostring = HTTP.urlencode })
	return values
end

local function setAuthority(comp, authority)
	comp.authority = authority
	comp.port = nil
	comp.host = nil
	comp.userinfo = nil
	comp.user = nil
	comp.password = nil

	authority = authority:gsub('^([^@]*)@', function(v)
		comp.userinfo = v
		return ''
	end)
	authority = authority:gsub("^%[[^%]]+%]", function(v)
		-- ipv6
		comp.host = v
		return ''
	end)
	authority = authority:gsub(':([^:]*)$', function(v)
		comp.port = tonumber(v)
		return ''
	end)
	if authority ~= '' and not comp.host then
		comp.host = authority:lower()
	end
	if comp.userinfo then
		local userinfo = comp.userinfo
		userinfo = userinfo:gsub(':([^:]*)$', function(v)
				comp.password = v
				return ''
		end)
		comp.user = userinfo
	end
	return authority
end

local function setQuery(comp, query)
	comp.query = HTTP.parse_query(query)
	return query
end

function HTTP.parse(url)
	local comp = {}
	setAuthority(comp, "")
    setQuery(comp, "")

	url = tostring(url or '')
	url = url:gsub('#(.*)$', function(v)
		comp.fragment = v
		return ''
	end)
	url = url:gsub('^([%w][%w%+%-%.]*)%:', function(v)
		comp.scheme = v:lower()
		return ''
	end)
	url = url:gsub('%?(.*)', function(v)
		setQuery(comp, v)
		return ''
	end)
	url = url:gsub('^//([^/]*)', function(v)
		setAuthority(comp, v)
		return ''
	end)
	comp.path = decode(url)

	return comp
end

function HTTP.url_query(url, query)
    return url .. '?' .. urlencode(query)
end

local FormData = {}
FormData.__index = FormData
HTTP.FormData = FormData

function FormData.New()
    local boundary = FormData.gen_boundary()
    return setmetatable({
        params = {},
        boundary = boundary,
    }, FormData)
end

function FormData:add_part(key, data, content_type, content_encoding, filename)
    content_type = content_type or 'application/octet-stream'

    local data_type = type(data)
    if data_type == 'table' then
        if content_type == HTTP.CT_JSON then
            data = JSON:encode(data)
        elseif content_type == HTTP.CT_URLENCODED then
            data = urlencode(data)
        end
    end
    data = tostring(data)
    content_encoding = content_encoding or 'binary'

    local params = self.params
    params[#params + 1] = {
        key = key,
        filename = filename,
        content_type = content_type,
        content_encoding = content_encoding,
        data = data,
    }
end

function FormData:add_simple_part(key, data)
    local params = self.params
    params[#params + 1] = {key = key, data = data}
end

function FormData:add_file(_key, _file)
    -- 实现一个高效的文件传输方式
end

local function encode_part(r, entry)
    r[#r + 1] = string.format('content-disposition: form-data; name="%s"', entry.key)

    if entry.filename then
        r[#r + 1] = string.format('; filename="%s"', entry.filename)
    end

    if entry.content_type then
        r[#r + 1] = "\r\ncontent-type: "
        r[#r + 1] = entry.content_type
    end

    if entry.content_encoding then
        r[#r + 1] = "\r\ncontent-transfer-encoding: %s"
        r[#r + 1] = entry.content_encoding
    end
    r[#r + 1] = "\r\n\r\n"
    r[#r + 1] = entry.data
    r[#r + 1] = "\r\n"
end

function FormData:content_type()
    return "multipart/form-data; boundary=" .. self.boundary
end

function FormData:build()
    local boundary = self.boundary
    local part = "--" .. boundary .. "\r\n"
    local r = {}

    for _, v in ipairs(self.params) do
        r[#r + 1] = part
        encode_part(r, v)
    end
    r[#r + 1] = "--"..boundary.."--"

    return table.concat(r)
end

function FormData.gen_boundary()
    local t = {"BOUNDARY-"}
    for i = 2,17 do
        t[i] = string.char(math.random(65, 90))
    end
    t[18] = "-BOUNDARY"
    return table.concat(t)
end

HTTP.CT_URLENCODED = 'application/x-www-form-urlencoded'
HTTP.CT_JSON = 'application/json'
HTTP.CT_FORMDATA = 'multipart/form-data'

function HTTP.smart_content(body, content_type)
    if type(body) == 'string' then
        return body
    elseif body == nil then
        return ''
    end

    assert(type(body) == 'table')

    if content_type == HTTP.CT_JSON then
        return JSON:encode(body)
    elseif content_type == HTTP.CT_URLENCODED then
        return urlencode(body)
    --这个留着
    --[[ elseif content_type == 'text/xml'  then
        return
    ]]
    end
    assert(false, content_type)
end

function HTTP.decode_by_content_type(acceptable, headers, body)
    local content_type = headers and headers['Content-Type']

    if body == nil or body == '' then
        return body
    end

    if content_type == nil then
        content_type = acceptable
    else
        content_type = content_type:match("^[^;]+")
    end

    if content_type ~= acceptable and acceptable ~= nil then
        return nil
    end

    if content_type == HTTP.CT_JSON then
        local succ, json = pcall(JSON.decode, JSON, body)
        if succ then
            return json
        else
            _ejoysdk.log("content_type error " .. json)
            _ejoysdk.log(body)
            return nil
        end
    end
    return body
end

function HTTP.get(url, params, cb)
    if params.httpdns or CONFIG.get_config('http_dns') then
        return HttpDns.get(url, params, cb)
    else
        return HTTP.process_get(url, params, cb)
    end
end

function HTTP.post(url, params, content_type, body, cb)
    if params.httpdns then
        return HttpDns.post(url, params, content_type, body, cb)
    else
        return HTTP.process_post(url, params, content_type, body, cb)
    end
end

local DnsCache = {}

local HttpDnsAdapter = {
    Adapter = 'Dnspod',
    Dnspod = {
        ttl = nil,   -- 需要ttl改成1
        api = "http://119.29.29.29/d"
    },
    Ejoy = {
        ttl = nil,   -- 需要ttl改成1
        api = "http://119.29.29.29"
    },
}

local function dnspod_read(data, ttl)
    local ret = {}
    local expires = nil
    if ttl then
        local entry = split_string(data, ",")
        data = entry[1]
        expires = tonumber(entry[2]) + os.time()
    end

    string.gsub(data, '[^;]+', function(c)
        if ttl then
            ret[#ret+1] = {c, expires}
        else
            ret[#ret+1] = c
        end
    end)

    if #ret == 0 then
        return
    end

    return ret
end

-- 这里不是很高效，因为其实很少需要这样搞的
local function dns_iterator(ips)
    if not ips then
        return function() end
    end

    local f, t, s = ipairs(ips)
    local now = os.time()
    local itr = function(t1, k_)
        local k, v = f(t1, k_)
        while k do
            local ip = nil
            if type(v) == 'string' then
                ip = v
            elseif v[2] <= now then
                ip = v[1]
            end
            if ip then
                return k, ip
            else
                k, v = f(t1, k)
            end
        end
    end
    return itr, t, s
end

local function make_dns_resp(ips, array)
    if array then
        local ret = {}
        for _, ip in dns_iterator(ips) do
            ret[#ret+1] = ip
        end
        if #ret ~= 0 then
            return ret
        end
    else
        local _, ip = next(dns_iterator(ips))
        return ip
    end
    return
end

local function httpdns_update_params(params, parsed_url)
    if not params.headers then
        params.headers = {}
    end
    local host = parsed_url.host
    if parsed_url.port then
        host = host .. ':' .. parsed_url.port
    end
    params.headers['Host'] = host
end

function HttpDnsAdapter.Dnspod:gethostbyname(domain, cb)
    local params = {
        dn = domain,
        ttl = self.ttl,
    }
    local url = HTTP.url_query(self.api, params)
    HTTP.process_get(url, {}, function(resp)
        if resp.status == 200 then
            if resp.body and #resp.body ~= 0 then
                return cb(dnspod_read(resp.body, self.ttl))
            end
        end
        return cb()
    end)
end

function HttpDnsAdapter.Dnspod:preload(domains, cb)
    local loop
    loop = function(rest, acc)
        local domain = table.remove(rest)
        if domain then
            self:gethostbyname(domain, function(ips)
                if ips then
                    acc[domain] = ips
                end
                loop(rest, acc)
            end)
        else
            return cb(acc)
        end
    end
    loop(domains, {})
end

function HttpDnsAdapter.Ejoy:gethostbyname(domain, cb)
    local params = {
        dn = domain,
        ttl = self.ttl,
    }
    local url = HTTP.url_query(self.api .. '/d', params)
    HTTP.process_get(url, {}, function(resp)
        if resp.status == 200 then
            if resp.body and #resp.body ~= 0 then
                return cb(dnspod_read(resp.body, self.ttl))
            end
        end
        return cb({})
    end)
end

local function ejoy_preload_read(data, _ttl)
    local ret = {}
    string.gsub(data, '[^\n]+', function(c)
        local entry = split_string(c, ":")
        local domain = entry[1]
        ret[domain] = dnspod_read(entry[2])
    end)
    return ret
end

function HttpDnsAdapter.Ejoy:preload(domains, cb)
    local params = {
        dn = domains:concat(','),
        ttl = self.ttl,
    }
    local url = HTTP.url_query(self.api .. '/q', params)
    HTTP.process_get(url, {}, function(resp)
        if resp.status == 200 then
            if resp.body and #resp.body ~= 0 then
                return cb(ejoy_preload_read(resp.body, self.ttl))
            end
        end
        return cb()
    end)
end

function HttpDns.is_numeric(domain)
    return string.find(domain, "^%d+%.%d+%.%d+%.%d+$")
end

function HttpDns.gethostbyname(domain, cb, return_array)
    if HttpDns.is_numeric(domain) then
        return cb(domain)
    end

    local cache = make_dns_resp(DnsCache[domain], return_array)
    if cache then
        return cb(cache)
    end

    local adapter = HttpDnsAdapter[HttpDnsAdapter.Adapter]
    adapter:gethostbyname(domain, function(ips)
        if ips then
            DnsCache[domain] = ips
            return cb(make_dns_resp(ips, return_array))
        else
            return cb()
        end
    end)
end

function HttpDns.prepare(url, params, cb)
    params.httpdns = nil
    local parsed_url = HTTP.parse(url)
    local host = parsed_url.host
    HttpDns.gethostbyname(host, function(ip)
        if ip then
            httpdns_update_params(params, parsed_url)
            url = url:gsub(host, ip, 1)
        end
        cb(url, params)
    end)
end

function HttpDns.get(url, params, cb)
    HttpDns.prepare(url, params, function(new_url, new_params)
        HTTP.process_get(new_url, new_params, cb)
    end)
end

function HttpDns.post(url, params, content_type, body, cb)
    HttpDns.prepare(url, params, function(new_url, new_params)
        HTTP.process_post(new_url, new_params, content_type, body, cb)
    end)
end

function HttpDns.preload(domains, cb)
    local adapter = HttpDnsAdapter[HttpDnsAdapter.Adapter]
    adapter:preload(domains, function(ret)
        for domain, ips in pairs(ret) do
            DnsCache[domain] = ips
        end
        if cb then
            cb()
        end
    end)
end

M.Utils = {}
M.Utils.split_string = split_string

function M.Sysinfo.os()
    return _ejoysdk.os()
end

function M.Sysinfo.bin_version()
    if _ejoysdk.bin_version then
        _ejoysdk.bin_version()
    else
        return "1.0.0"
    end
end

local LazyKeyStore = Class:Inherit("LazyKeyStore")

function LazyKeyStore:_init(key, no_auto_save, is_json)
    self.is_json = is_json
    self.key = key
    self.value = nil
    self.auto_save = (not no_auto_save)
end

function LazyKeyStore:set(value)
    self.value = value
    if self.auto_save then
        self:save()
    end
end

function LazyKeyStore:save()
    local value = self.value
    if self.is_json then
        value = JSON:encode(value)
    end
    M.KeyStore.set(self.key, value)
end

function LazyKeyStore:get()
    if not self.value then
        self.value = M.KeyStore.get(self.key)
        if self.is_json and self.value then
            self.value = JSON:decode(self.value)
        end
    end
    return self.value
end

function LazyKeyStore:delete()
    self.value = nil
    M.KeyStore.delete(self.key)
end

M.LazyKeyStore = LazyKeyStore

return M
