--
-- User: sean
-- Date: 15-12-1
-- Time: 下午3:45
-- API 尽量多的检查参数错误， 给用户更多方便
--

local E = require "ejoysdk_lua.ejoysdk"
local EV = require "ejoysdk_lua.ejoysdk_vendors"
local ER = require "ejoysdk_lua.ejoysdk_resource"

local ET = require "ejoysdk_lua.ejoysdk_topic"

local HTTP = E.HTTP

local M = {}

-- 用以记录当前登录的用户信息
local UserInfoMt = {
    game = nil,
    server = nil,
    region = nil,
    token = nil,
    uid = nil,
    platform = nil,
    ptoken = nil,
    pid = nil,
    guest = nil,
    with = nil,
    with_account = nil,
    ext = nil,
}
UserInfoMt.__index = UserInfoMt

local login_listener = nil
local pay_listener = nil
local bind_listener = nil
local queue_listener = nil

local function new_user_info(tbl)
    tbl = tbl or {}
    return setmetatable(tbl, UserInfoMt)
end

local USER_INFO = new_user_info()

local EJOY_TOKEN = E.LazyKeyStore:New('EJOY_TOKEN')
local VENDOR_LOGIN = E.LazyKeyStore:New('VENDOR_LOGIN', false, true)
local LOGIN_INFO = E.LazyKeyStore:New("LOGIN_INFO", true, true)

local ON_QUEUE_LOOP = nil

function M.gangplank_url(api, ver)
    local product = E.CONFIG.get_config('product'):lower()
    local url_base = E.CONFIG.get_config('gangplank')

    if ver then
        api = '/v' .. ver .. api
    end

    return url_base .. '/gp/' .. product .. api
end

local APIS = {
    login = true,
    queue = true,
    queue_dropout = true,
    create_order = true,
    bind = true,
    query = true,
    product_infos = true,
    access = true,
}

local GangplankUrl = {
    __index = function(self, key)
        local ver = rawget(self, 'ver')
        local apis = rawget(self, 'apis') or APIS
        assert(apis[key], "api not found: " .. key)
        return M.gangplank_url('/' .. key, ver)
    end,
    new = function(self, ver, apis)
        return setmetatable({
            ver = ver,
            apis = apis,
        }, self)
    end
}

local url_tbl_v1 = GangplankUrl:new()
local url_tbl_v2 = GangplankUrl:new('2')

local function _gangplank_v1_post(cmd, params, cb)
    local url = url_tbl_v1[cmd]
    assert(url, "gangplank api: " .. cmd .. " not found")

    HTTP.post(url, {acceptable=HTTP.CT_JSON}, HTTP.CT_URLENCODED , params, function(resp)
        cb(resp.status, resp.body)
    end)
end

local function gangplank_v2_post(cmd, params, cb)
    local url = url_tbl_v2[cmd]
    assert(url, "gangplank api: " .. cmd .. " not found")

    HTTP.post(url, {acceptable=HTTP.CT_JSON}, HTTP.CT_JSON, params, function(resp)
        if resp.status == 200 then
            if resp.body.code == 0 then
                cb(200, resp.body)
            else
                cb(resp.body.code, resp.body)
            end
        else
            cb(resp.status, resp.body)
        end
    end)
end

local gangplank_post = gangplank_v2_post

local function login_base(server, region, outsource, token, cb)
    assert(type(server) == 'string' and server ~= '', 'server should be string')
    region = region or E.CONFIG.get_config('product')
    assert(type(region) == 'string', 'region should be string')
    assert(type(outsource) == 'table', 'outsource should be table')
    if outsource.platform  then
        assert(type(outsource.ptoken) == 'string', 'platform set, outsource.ptoken should be string')
    end
    assert(type(token) == 'string' or token == nil, 'token should be string')
    assert(type(cb) == 'function', 'cb should be function')

    local secret = _ejoysdk_crypt.randomkey()

    local params = {
        secret = _ejoysdk_crypt.base64encode(_ejoysdk_crypt.dhexchange(secret)),
        game = E.CONFIG.get_config('product'),
        server = server,
        region = region,
        token = token,
        platform = outsource.platform,
        ptoken = outsource.ptoken,
        pid = outsource.pid,
        guest = outsource.guest,
        with = outsource.with,
        with_account = outsource.with_account,
        ext = outsource.ext,
        appname = outsource.appname,
        pkg_info = {
            os = E.CONFIG.get_config('os'),
            tag = E.CONFIG.get_config('unisdk_meta').tag,
            uuid = E.Sysinfo.idfa(),
        }
    }

    gangplank_post('login', params, function(status, body)
        USER_INFO = new_user_info()

        if status == 200 or status == 413 then
            USER_INFO = new_user_info({
                game = params.game,
                server = params.server,
                region = params.region,
                token = body.token,
                uid = body.uid,
                platform = params.platform,
                ptoken = params.ptoken,
                ext = params.ext,
                pid = body.pinfo.pid,
                guest = params.guest,
                with = body.pinfo.with,
                with_account = body.pinfo.with_account,
            })
        end

        if status == 200 then
            local server_secret = _ejoysdk_crypt.base64decode(body.server_secret)
            local real_secret = _ejoysdk_crypt.dhsecret(server_secret, secret)
            cb(true, body.token, body.game_token, real_secret, body.pinfo)
        else
            cb(false, status, body, secret)
        end
    end)
end

local function call_queue_listener(status, ...)
    if queue_listener then
        queue_listener(status, ...)
    end
end

local function get_interval(queue)
    if queue > 50 then
        return 5
    elseif queue > 5 then
        return 3
    else
        return 1
    end
end

local function queue_loop(ticket, queue, secret, cb)
    local interval = get_interval(queue)

    local timer_cb = function()
        if not ON_QUEUE_LOOP then
            return
        end
        local params = {
            ticket = ticket,
        }
        gangplank_post('queue', params, function(status, body)
            if not ON_QUEUE_LOOP then
                return
            end
            if status == 200 then
                local server_secret = _ejoysdk_crypt.base64decode(body.server_secret)
                local real_secret = _ejoysdk_crypt.dhsecret(server_secret, secret)

                call_queue_listener('end')
                ON_QUEUE_LOOP = nil
                cb(true, body.game_token, real_secret, USER_INFO.pid)
                ET.publish('login', USER_INFO)
            elseif status == 413 then
                local new_queue_count = body.queue
                call_queue_listener('loop', new_queue_count)
                queue_loop(ticket, new_queue_count, secret, cb)
            else
                call_queue_listener('end')
                ON_QUEUE_LOOP = nil
                cb(false, status)
            end
        end)
    end
    E.Timer.once(interval, timer_cb)
end


local function login_callback_builder(region, outsource, info, cb)
    local guest = outsource == nil

    local update_token = function(token, pinfo)
        if guest then
            EJOY_TOKEN:set(token)
            info = pinfo
        else
            local vendor_name = EV.get_vendor_name(outsource)
            local vendor = EV.get(vendor_name)
            info = vendor.merge_info(info, pinfo)

            outsource.pid = info.pid
            VENDOR_LOGIN:set({
                region = region,
                token = token,
                outsource = outsource,
                info = info
            })
        end
        return info
    end

    return function(succ, token, game_token, secret, pinfo)
        if succ then
            local updated_info = update_token(token, pinfo)
            cb(true, game_token, secret, updated_info.pid)
            ET.publish('login', USER_INFO)
        else
            local status = token
            local body = game_token

            _ejoysdk.log("login failed " .. tostring(status))

            if status == 401 then
                if guest then
                    EJOY_TOKEN:delete()
                else
                    VENDOR_LOGIN:delete()
                end
            elseif status == 413 then       -- 排队
                update_token(body.token, body.pinfo)
                local queue = body.queue
                call_queue_listener('start', queue)
                ON_QUEUE_LOOP = body.ticket
                return queue_loop(body.ticket, queue, secret, cb)
            elseif status == 406 or status == 462 or status == 407 then
                VENDOR_LOGIN:delete()
            end
            cb(false, status)
        end
    end
end

--[
-- vendor_login 使用已经登录过的帐户 登录游戏
--
-- @param server 要登录的服务器, 名字与在平台监听（mq routing key）的名字相同
-- @param region 账号归属于哪个区域，gangplank 的一个帐号, 在一个 region 下是唯一,
-- 一个游戏建议只设置一个统一的 region, 不设置 region, region 会使用产品代号
--]
function M.vendor_login(server, region)
    assert(login_listener, "register login listener first")
    assert(type(server) == 'string', 'server should be string')
    region = region or E.CONFIG.get_config('product')
    assert(type(region) == 'string', 'region should be string')

    local history = VENDOR_LOGIN:get()
    assert(history, "vendor login history not found")

    local vendor_name = EV.get_vendor_name(history.outsource)
    local vendor = EV.get(vendor_name)
    assert(vendor, 'vendor: ' .. tostring(vendor_name) .. 'not found')

    local token = history.token
    if region ~= history.region then
        token = nil
    end

    if vendor.simple_token() then
        login_base(
            server, region, history.outsource, token,
            login_callback_builder(region, history.outsource, history.info, login_listener)
        )
    else
        LOGIN_INFO:set({
            type = 'login',
            server = server,
            region = region,
            token = token,
            vendor = vendor_name,
        })
        vendor.check_token(history.outsource, history.info)
    end
end

function M.vendor_logout()
    local history = VENDOR_LOGIN:get()
    assert(history, "vendor login history not found")

    local vendor_name = EV.get_vendor_name(history.outsource)
    local vendor = EV.get(vendor_name)
    assert(vendor, 'vendor: ' .. tostring(vendor_name) .. 'not found')

    vendor.logout(function()
        VENDOR_LOGIN:delete()
        USER_INFO = new_user_info()
    end)
end

--[
-- login 登录游戏
--
-- @param server 要登录的服务器, 名字与在平台监听（mq routing key）的名字相同
-- @param region 账号归属于哪个区域，gangplank 的一个帐号, 在一个 region 下是唯一,
-- 一个游戏建议只设置一个统一的 region, 不设置 region, region 会使用 server
-- @param vendor_name 要用来登录的第三方平台, nil 表示游客登录
--]
function M.login(server, region, vendor_name)
    assert(login_listener, "register login listener first")
    assert(type(server) == 'string', 'server should be string')
    region = region or E.CONFIG.get_config('product')
    assert(type(region) == 'string', 'region should be string')

    if vendor_name == nil then
        -- 纯游客登录
        local token = EJOY_TOKEN:get()
        login_base(
            server, region, {}, token,
            login_callback_builder(region, nil, nil, login_listener)
        )
    else
        local vendor = EV.get(vendor_name)
        assert(vendor, 'vendor: ' .. tostring(vendor_name) .. 'not found')

        LOGIN_INFO:set({
            type = 'login',
            server = server,
            region = region,
            token = nil,
            vendor = vendor_name,
        })

        vendor.login()
    end
end


function M.cancel_queue()
    if ON_QUEUE_LOOP then
        local params = {
            ticket = ON_QUEUE_LOOP,
        }
        gangplank_post('queue_dropout', params, function() end)
        call_queue_listener('end')
        ON_QUEUE_LOOP = nil
    end
end

local function bind_base(token, outsource, force, cb)
    local params = {
        token = token,                          -- 选择绑定的平台token
        game = E.CONFIG.get_config('product'),
        server = USER_INFO.server,
        region = USER_INFO.region,
        platform = outsource.platform,
        ptoken = outsource.ptoken,
        pid = outsource.pid,
        ext = outsource.ext,
        force = force,
    }

    gangplank_post('bind', params, function(status, body)
        if status == 200 then
            cb.success(body)
        elseif status == 403 then
            cb.conflict(body.conflict)
        else
            cb.error(status)
        end
    end)
end

local function bind_callback_builder(outsource, info)
    local guest = USER_INFO.platform == nil

    local token
    if guest then
        token = EJOY_TOKEN:get()
    else
        token = VENDOR_LOGIN:get().token
    end

    local cb = {
        success = function(body)
            if guest then
                EJOY_TOKEN:delete()
            end
            VENDOR_LOGIN:set({
                region = USER_INFO.region,
                token = body.token,
                outsource = outsource,
                info = info
            })
            bind_listener("succ", body.token)
        end,
        error = function(status)
            if status == 401 then
                if guest then
                    EJOY_TOKEN:delete()
                else
                    VENDOR_LOGIN:delete()
                end
            end
            bind_listener("error", status)
        end,
        conflict = function(conflict)
            local continure_bind = function(selected_token)
                if selected_token == token then
                    bind_base(
                        token, outsource, true,
                        bind_callback_builder(outsource, info))
                end
            end
            bind_listener("conflict", conflict, continure_bind)
        end
    }
    return cb
end

local function bind(vendor_name)
    local vendor = EV.get(vendor_name)
    assert(vendor, 'vendor: ' .. tostring(vendor_name) .. 'not found')

    LOGIN_INFO:set({
        type = 'bind',
        token = USER_INFO.token,
        vendor = vendor_name,
    })

    vendor.login()
end

local function query(server, region, outsource, cb)
    assert(type(server) == 'string' and server ~= '', 'server should be string')
    assert(type(region) == 'string' or region == nil, 'region should be string or nil')
    assert(type(outsource.platform) == 'string', 'outsource.platform should be string')
    assert(type(outsource.ptoken) == 'string', 'outsource.ptoken should be string')
    assert(type(outsource.pid) == 'string' or outsource.pid == nil, 'outsource.pid should be string or nil')

    assert(type(cb.success) == 'function', 'cb.success should be function')
    assert(type(cb.error) == 'function', 'cb.error should be function')

    local params = {
        game = E.CONFIG.get_config('product'),
        server = server,
        region = region,
        platform = outsource.platform,
        ptoken = outsource.platform,
        pid = outsource.pid,
        ext = outsource.ext,
    }

    gangplank_post('query', params, function(status, body)
        if status == 200 then
            cb.success(body.tokens)
        else
            cb.error(status)
        end
    end)
end

ET.subscribe("app_on_stop", function()
    if LOGIN_INFO:get() then
        LOGIN_INFO:save()
    end
end)

local function create_order_base(type_, amount, channel, outsource, cb, override)
    assert(type(type_) == 'string' and type_ ~= '', 'type should be string')
    assert(type(amount) == 'number' and amount > 0, 'amount should be number and larger then 0')
    assert(type(cb) == 'function', 'cb should be function')
    assert(USER_INFO.token, "need login")

    local params = {
        ["type"] = type_,
        amount = amount,
        platform = outsource.platform,
        ptoken = outsource.ptoken,
        pid = outsource.pid,
        with = outsource.with,
        with_account = outsource.with_account,
        ext = outsource.ext,
        channel = channel,
        game = E.CONFIG.get_config('product'),
        server = USER_INFO.server,
        token = USER_INFO.token,
    }

    if override then
        for k, v in pairs(override) do
            params[k] = v
        end
    end

    gangplank_post('create_order', params, function(status, body)
        if status == 200 then
            cb(true, body.order_id, body)
        else
            cb(false)
        end
    end)
end

function M.pay(channel, product_id, count, override)
    assert(pay_listener, "register pay listener first")

    local vendor = EV.get(channel)
    assert(vendor, 'vendor: ' .. tostring(vendor) .. 'not found')

    if not vendor.can_pay()  then
        ET.publish('purchased', false, ER.order.CANT_PURCHASE)
        return
    end

    assert(vendor.product_list()[product_id], "product " .. product_id .. " not found")

    local outsource = {
        platform = USER_INFO.platform,
        ptoken = USER_INFO.ptoken,
        pid = USER_INFO.pid,
        with = USER_INFO.with,
        with_account = USER_INFO.with_account,
        ext = USER_INFO.ext,
    }

    local channel_target = USER_INFO.with or USER_INFO.platform
    if channel == channel_target then
        channel = nil
    end

    create_order_base(product_id, count, channel, outsource, function(succ, order_id, body)
        if succ then
            vendor.pay(product_id, count, order_id, body)
        else
            pay_listener(false)
        end
    end, override)
end

local function product_infos_base(channel, cb)
    assert(USER_INFO.token, "need login")

    local params = {
        token = USER_INFO.token,
        game = E.CONFIG.get_config("product"),
        channel = channel,
    }
    gangplank_post('product_infos', params, function(status, body)
        if status == 200 then
            local product_infos = {}

            for _, v in ipairs(body.product_infos) do
                product_infos[v.product_id] = v
            end

            cb(true, product_infos)
        else
            cb(false)
        end
    end)
end

M.VENDOR_LOGIN_ERROR = -1

-- 所有的base接口，都是不负责和第三方sdk通讯的接口
-- 不带base的接口，都是打算以后做一站式服务的接口.

M.login_base = login_base
M.bind_base = bind_base
M.product_infos_base = product_infos_base
M.create_order_base = create_order_base

-- 尽量不用这些接口 begin:
function M.set_user_info(info)
    USER_INFO = info
end
M.get_ejoy_token = function()
    return EJOY_TOKEN:get()
end
-- end

M.get_vendor_auth_info = function()
    return VENDOR_LOGIN:get()
end
M.offline = function()
    USER_INFO = new_user_info()
end
M.bind = bind
M.query = query

function M.user_info()
    return USER_INFO
end

function M.access(cb)
    local params = {
        token = USER_INFO.token,
    }
    gangplank_post('access', params, function(status, body)
        if(status == 200) then
            cb(true, body)
        else
            cb(false, body)
        end
    end)
end

--[
-- 有一些SDK会在程序启动的时候直接推送消息给我们的程序，告知登录信息，
-- 只能用如此复杂的登录/付费流程
--
-- @param opts: 一个 table，以 VENDOR 名字为key， 比如 EJOY, WEIXIN, ALIPAY,
-- 值为一个 table， 会传递给对应的 VENDOR 作为初始化参数, 比如
-- {EJOY = {hide_email_login = true}, ALIPAY={}}, 没有出现的 VENDOR 不会被初始化
--
-- @param listeners: 一个 table， 包含四个 key, auth_listener, pay_listener, bind_listener, queue_listener
--]

function M.init(opts, listeners)
    login_listener = listeners.auth_listener
    pay_listener = listeners.pay_listener
    bind_listener = listeners.bind_listener
    queue_listener = listeners.queue_listener

    EV.init(opts, {
        auth_listener = function(vendor, succ, outsource, info)
            local login_info = LOGIN_INFO:get()
            if login_info and vendor == login_info.vendor then

                LOGIN_INFO:delete()

                if login_info.type == 'login' then
                    if succ then
                        local region = login_info.region
                        login_base(
                            login_info.server, region, outsource, login_info.token,
                            login_callback_builder(region, outsource, info, login_listener)
                        )
                    else
                        VENDOR_LOGIN:delete()
                        if outsource then
                            outsource.type = 'vendor_login_failed'
                            outsource.vendor = vendor
                            ET.publish('trace_event', 'boot', outsource)
                        end
                        login_listener(false, M.VENDOR_LOGIN_ERROR, outsource)
                    end
                elseif login_info.type == 'bind' then
                    if succ then
                        bind_base(
                            login_info.token, outsource, false,
                            bind_callback_builder(outsource, info)
                        )
                    else
                        bind_listener("error", M.VENDOR_LOGIN_ERROR, outsource)
                    end
                end
            end
        end,
        pay_listener = function(_vendor, succ, order_id, ext)
            pay_listener(succ, order_id, ext)
        end,
        switch_listener = function(_vendor, server, region, outsource, info)
            LOGIN_INFO:delete()
            login_base(
                server, region, outsource, nil,
                login_callback_builder(region, outsource, info, login_listener)
            )
        end
    })
end

return M
