local E = require "ejoysdk_lua.ejoysdk"
local ETOPIC = require "ejoysdk_lua.ejoysdk_topic"
local ER = require 'ejoysdk_lua.ejoysdk_resource'
local EG = require "ejoysdk_lua.ejoysdk_gangplank"

local M = {}

M.HOLO_TOPIC = 'holo'

M.GENDER_UNKNOWN = 0
M.GENDER_MALE = 1
M.GENDER_FAMALE = 2

local inited = false
local info_cache = nil

local URLS = {}
local url_items = {
    info = 'user/info',
    user_infos='user/user_infos',
    login_token = 'user/login_token',
    location = 'user/info/location',
    login_cb = 'login_cb',
    set_avatar = 'user/info/photo/show/avatar',
    transform = 'audio/client/transform',
    audiometa = 'audio/client/text',
}

local function require_params()
    local token = EG.user_info().token

    return {
        acceptable = E.HTTP.CT_JSON,
        headers = {['Ejoy-Token']= token}
    }
end

local login_handler
do
    login_handler = function()
        M.get_info(function(succ)
            ETOPIC.unsubscribe(login_handler)
            ETOPIC.publish('holo_inited', succ)
            if succ then
                inited = true
            end
        end)
    end
end

do
    local product = E.CONFIG.get_config("product"):lower()
    local url_prefix = E.CONFIG.get_config("holo") .. '/holo/' .. product .. '/api/1/'
    for k, v in pairs(url_items) do
        URLS[k] = url_prefix .. v
    end
end

function M.init()
    if inited then
        return
    end
    ETOPIC.subscribe('login', login_handler)
end

local function update_info(new_info)
    new_info.code = nil
    new_info.message = nil
    info_cache = new_info
end

function M.get_info(cb)
    E.HTTP.get(URLS.info, require_params(), function(resp)
        if resp.status == 200 then
            update_info(resp.body)
            cb(true, info_cache)
        else
            cb(false)
        end
    end)
end

function M.set_info(name, gender, birthday, bio, cb)
    local params = {
        name = name,
        gender = gender,
        birthday = birthday,
        bio = bio,
    }
    E.HTTP.post(URLS.info, require_params(), E.HTTP.CT_JSON, params, function(resp)
        if resp.status == 200 then
            update_info(resp.body)
            cb(true, info_cache)
        else
            cb(false)
        end
    end)
end

function M.update_location(longitude, latitude, cb)
    local params = {
        longitude = longitude,
        latitude = latitude,
    }
    E.HTTP.post(URLS.location, require_params(), E.HTTP.CT_JSON, params, function(resp)
        if resp.status == 200 then
            update_info(resp.body)
            cb(true, info_cache)
        else
            cb(false)
        end
    end)
end

-- 没有特殊情况，用这个接口
function M.info()
    return info_cache
end

function M.user_infos(ids, cb)
    assert(type(ids) == 'table', 'ids should be string array')
    assert(#ids > 0 and #ids <= 30, 'ids amount should not larger then 30')

    local params = {
        ids = ids
    }
    E.HTTP.post(URLS.user_infos, require_params(), E.HTTP.CT_JSON, params, function(resp)
        if resp.status == 200 then
            local body = resp.body
            if body.code == 0 then
                cb(true, body.infos)
                return
            end
        end
        cb(false)
    end)
end

function M.set_avatar(source, id, cb)
    local params = {
        source = source,
        id = id,
    }

    E.HTTP.post(URLS.set_avatar, require_params(), E.HTTP.CT_JSON, params, function(resp)
        if resp.status == 200 then
            local body = resp.body
            if body.code == 0 then
                update_info(resp.body)
                cb(true, info_cache)
                return
            end
        end
        cb(false)
    end)
end

local PHOTO_THUMB = 0
local PHOTO_ORIGIN = 1

local image_source = {
    MOMO= function(t, info)
        local prefix = 'http://img.immomo.com/album/'

        local id = info.id
        local dir1 = id:sub(1, 2)
        local dir2 = id:sub(3, 4)

        local suffix = '_S.jpg'
        if t == PHOTO_ORIGIN then
            suffix = '_L.jpg'
        end
        return prefix .. dir1 .. '/' .. dir2 .. '/' .. id .. suffix
    end,

    -- 建议直接游戏内嵌
    BI = function(_t, _info)
        return
    end,

    HOLO = function(t, info)
        local product = E.CONFIG.get_config("product"):lower()
        local prefix = E.CONFIG.get_config("holo-cdn") .. '/' .. product .. '/image/'

        local id = info.id
        local dir1 = id:sub(-4, -3)
        local dir2 = id:sub(-2)

        local suffix = '_t.jpg'
        if t == PHOTO_ORIGIN then
            suffix = '_o.' .. info.type
        end
        return prefix .. dir1 .. '/' .. dir2 .. '/' .. id .. suffix
    end,

    CHANGBA = function (_t, info)
        return info.id
    end,

    FB = function (_t, info)
        return info.id
    end,
}

local audio_source = {

    HOLO = function(id, type_)
        if not type_ then
            type_ = 'amr'
        end
        local product = E.CONFIG.get_config("product"):lower()
        local prefix = E.CONFIG.get_config("holo-cdn") .. '/' .. product .. '/audio/'

        local dir1 = id:sub(-4, -3)
        local dir2 = id:sub(-2)

        return string.format('%s%s/%s/%s.%s', prefix, dir1, dir2, id, type_)
    end,

    RECORD_DIR = function(id, type_)
        if not type_ then
            type_ = 'amr'
        end

        local product = E.CONFIG.get_config("product"):lower()
        local prefix = E.CONFIG.get_config("holo-cdn") .. '/' .. product .. '/audio/'

        local dir1 = id:sub(-4, -3)
        local dir2 = id:sub(-2)

        return string.format('%s%s/%s/%s.%s', prefix, dir1, dir2, id, type_)
    end
}

function M.get_thumb_url(photo_info)
    local f = image_source[photo_info.source]
    return f(PHOTO_THUMB, photo_info)
end

function M.get_origin_url(photo_info)
    if not photo_info then
        return
    end
    local f = image_source[photo_info.source]
    return f(PHOTO_ORIGIN, photo_info)
end

function M.get_audio_url(id, type_)
    -- 目前只有 HOLO 一种
    local f = audio_source.HOLO
    return f(id, type_)
end

local function open_url(cb_url, is_webview, login_token)
    local url = E.HTTP.url_query(URLS.login_cb, {login_token = login_token, cb=cb_url})
    if is_webview then
        E.WebView.open(url)
    else
        E.Sysinfo.open_url(url)
    end
end

function M.edit(is_webview, package)
    local product = E.CONFIG.get_config("product"):lower()
    local cb_url = E.CONFIG.get_config("holo") .. '/holo/' .. product .. '/'

    if package then
        cb_url = cb_url .. '#!/home?package=' .. package
    end

    E.HTTP.get(URLS.login_token, require_params(), function(resp)
        if resp.status == 200 then
            local body = resp.body
            if body.code == 0 then
                local token = body.login_token
                open_url(cb_url, is_webview, token)
                return
            end
        end
        ETOPIC.publish(M.HOLO_TOPIC, ER.holo.EDITOR_ERROR)
    end)
end

function M.open_login_url(url, is_webview)
    E.HTTP.get(URLS.login_token, require_params(), function(resp)
        if resp.status == 200 then
            local body = resp.body
            if body.code == 0 then
                local token = body.login_token
                open_url(url, is_webview, token)
                return
            end
        end
        ETOPIC.publish(M.HOLO_TOPIC, ER.holo.WEB_LOGIN_ERROR)
    end)
end

function M.speech_recognize(opt, cb)
    local formdata = E.HTTP.FormData.New()
    formdata:add_simple_part('type', opt.format)

    local mime
    if opt.format == 'amr' then
        formdata:add_simple_part('rate', '8000')
        mime = 'audio/amr'
    else
        mime = 'audio/m4a'
    end
    formdata:add_part('audio', opt.bytes, mime, 'binary', 'audio' .. opt.format)

    local content_type = formdata:content_type()
    local build = formdata:build()
    local post_cb = function(resp)
        if resp.status ~= 200 then
            cb({succ=false, message='http error: ' .. resp.status, text=''})
            return
        end

        local body = resp.body
        if body.code ~= 0 then
            cb({succ=false, message=body.err_msg, text=''})
            return
        end

        local ret = {
            succ=true,
            text=body.text,
            id=body.id,
            sign=body.sign,
            type=body.type,
            duration=body.duration,
        }
        cb(ret)
    end

    E.HTTP.post(URLS.transform, require_params(), content_type, build, post_cb)
end

return M
