-- iOS 平台相关的ejoysdk lua 代码
local JSON = require "ejoysdk_lua.ejoysdk_json"
local ET = require "ejoysdk_lua.ejoysdk_topic"

local M = {}
local cbs = {}
local cb_id = math.random(1000, 9999)

M.PLATFORM = {
    OS = "iOS",
    HTTP_UA = "AFNetworking/2.4.5 (iOS; iPhone)"
}


function M.async_call(fn_name, cb, ...)
    local id = cb_id
    cb_id = cb_id + 1
    cbs[id] = cb

    _ejoysdk[fn_name](id, ...)
    return id
end

function M.async_call_with_opts(fn_name, opts, cb, ...)
    local id = M.async_call(fn_name, cb, ...)

    if opts and opts.timeout and opts.timeout_cb then
        M.Timer.once(opts.timeout, function()
            if cbs[id] then
                cbs[id] = nil
                opts.timeout_cb()
            end
        end)
    end
end

local function async_callback(id, ...)
    local cb = cbs[id]
    if cb then
        cbs[id] = nil
        cb(...)
    end
end

_ejoysdk.register_cb("ASYNC_CALL", async_callback)

function M.log(params)
    local log = JSON:encode(params)
    return _ejoysdk.log(log)
end
function M.printl(content)
    M.log({msg= content})
end

local HTTP_EVENT = 'HTTP_EVENT'
local http_progress_cbs = {}

_ejoysdk.register_cb(HTTP_EVENT, function(cbid, received, total)
    local params  = http_progress_cbs[cbid]
    if params then
        if received == -1 then
            http_progress_cbs[cbid] = nil
        else
            params.progress(params.url, params.file, received, total)
        end
    end
end)

local HTTP = {}
M.HTTP = HTTP

function HTTP.process_get(url, params, cb)
    local headers = HTTP.check_and_update_headers(params.headers)
    local name = 'http_get'
    if params.file then
        name = 'http_get_file'
    end
    local cbid = M.async_call_with_opts(name, {
        timeout=params.timeout,
        timeout_cb=function() cb({status=0}) end
    }, function(status, resp_headers, body)
        resp_headers = HTTP.Header.New(resp_headers)
        if not params.raw_body then
            body = HTTP.decode_by_content_type(params.acceptable, resp_headers, body)
        end
        cb({
            status=status,
            headers=resp_headers,
            body=body,
        })
    end, url, headers, params.file)
    if params.file and params.progress then
        http_progress_cbs[cbid] = {url=url, file=params.file, progress=params.progress}
    end
end

function HTTP.process_post(url, params, content_type, body, cb)
    local headers = HTTP.check_and_update_headers(params.headers)
    if headers['Content-Type'] then
        assert(headers['Content-Type'] == content_type)
    end

    body = HTTP.smart_content(body, content_type)

    M.async_call_with_opts('http_post', {
        timeout=params.timeout,
        timeout_cb=function() cb({status=0}) end
    }, function(status, resp_headers, resp_body)
        resp_headers = HTTP.Header.New(resp_headers)
        if not params.raw_body then
            resp_body = HTTP.decode_by_content_type(params.acceptable, resp_headers, resp_body)
        end
        cb({
            status=status,
            headers=resp_headers,
            body=resp_body,
        })
    end, url, headers, body, content_type)
end

local KEYSTORE_KEY = 'ejoysdk'
local KeyStore = {}
M.KeyStore = KeyStore

function KeyStore.get(key)
  local store = _ejoysdk.load_from_keychain(KEYSTORE_KEY) or {}
  return store[key]
end

function KeyStore.set(key, value)
  local store = _ejoysdk.load_from_keychain(KEYSTORE_KEY) or {}
  store[key] = value
  _ejoysdk.save_to_keychain(KEYSTORE_KEY, store)
end

function KeyStore.delete(key)
  local store = _ejoysdk.load_from_keychain(KEYSTORE_KEY)
  if store then
    store[key] = nil
    _ejoysdk.save_to_keychain(KEYSTORE_KEY, store)
  end
end

function KeyStore.clear()
  _ejoysdk.delete_from_keychain(KEYSTORE_KEY)
end

local WebView = {}
M.WebView = WebView

function WebView.open(url, injection, option)
    local injectionString = '{".ejoy.com": {}}'
    if injection then
        if not injection['.ejoy.com'] then
            injection['.ejoy.com'] = {}
        end
        injectionString = JSON:encode(injection)
    end
    local optionString = '{}'
    if option then
        optionString = JSON:encode(option)
    end
    return _ejoysdk.webview_open(url, injectionString, optionString)
end

function WebView.close()
    return _ejoysdk.webview_close()
end

function WebView.is_opened()
    return _ejoysdk.webview_is_opened()
end

local WEBVIEW_EVENT = 'WEBVIEW_EVENT'
local WEBVIEW_JSARGS_EVENT = 0;
local WEBVIEW_CLOSE_EVENT = 1;

_ejoysdk.register_cb(WEBVIEW_EVENT, function(cbid, value)
    value = JSON:decode(value)
    if cbid == WEBVIEW_JSARGS_EVENT then
        if value.args.type == 'oauthUri' then
            ET.publish('logindone', value);
            return
        end
    end
    if cbid == WEBVIEW_CLOSE_EVENT then
        ET.publish('webview_close', value);
        return
    end
end)

local APP_EVENT = 'APP_EVENT'
local APP_OPENURL_EVENT = 0
local APP_ON_STOP_EVENT = 1

local _app_event_handler = {
    [APP_OPENURL_EVENT] = function(json_str)
        local obj = JSON:decode(json_str)
        ET.publish('urlopen', obj.url)
    end,
    [APP_ON_STOP_EVENT] = function()
        ET.publish("app_on_stop")
    end,
}
_ejoysdk.register_cb(APP_EVENT, function(cbid, ...)
    local handler = _app_event_handler[cbid]

    if handler then
        handler(...)
    end
end)

local Modal = {}
M.Modal = Modal

function Modal.open(title, option, cb)
    local optionString = '{}'
    if option then
        optionString = JSON:encode(option)
    end
    cb = cb or function() end
    M.async_call('modal_open', cb, title, optionString)
end

function Modal.alert(title, message, cb)
    local option = {
        message= tostring(message),
        buttons= {'确定'}
    }
    local optionString = JSON:encode(option)
    local cb_wrap = function(_index) cb() end
    M.async_call('modal_open', cb_wrap, title, optionString)
end

function Modal.confirm(title, message, cb)
    local option = {
        message= tostring(message),
        buttons= {'取消', '确定'}
    }
    local optionString = JSON:encode(option)
    local cb_wrap = function(index)
        cb(index ~= 0);
    end
    M.async_call('modal_open', cb_wrap, title, optionString)
end

local Toast = {}
M.Toast = Toast

function Toast.show(message, option)
    local optionString = '{}'
    if option then
        optionString = JSON:encode(option)
    end
    return _ejoysdk.toast_open(message or '', optionString)
end

function Toast.hide()
    return _ejoysdk.toast_open('', '{}')
end

local Sysinfo = {}
M.Sysinfo = Sysinfo

function Sysinfo.idfa()
    return _ejoysdk.sysinfo_idfa()
end

function Sysinfo.open_url(url)
    return _ejoysdk.sysinfo_open_url(url)
end

local Timer = {}
M.Timer = Timer

function Timer.once(interval, cb)
    M.async_call('timer_once', cb, interval)
end

local File = {}
M.File = File

function File.writefile(filename, filedata, append)
    local append_int = 0
    if append == true then
        append_int = 1
    end
    return _ejoysdk.writefile(filename, filedata, append_int)
end

local Media = {
    DEFAULT_MAX_FILESIZE = 1024 * 1024 * 5,
    DEFAULT_MAX_DURATION = 60 * 1000 * 3,
    DEFAULT_SAMPLING_RATE = 8000,
    DEFAULT_ENCODING_BIT_RATE = 16,
    DEFAULT_CHANNEL = 1,
}
M.Media = Media

function Media.start_record(opt, cb)
    opt = opt or {}

    local max_filesize = Media.DEFAULT_MAX_FILESIZE
    if opt.max_filesize and opt.max_filesize <= max_filesize then
        max_filesize = opt.max_filesize
    end
    local max_duration = Media.DEFAULT_MAX_DURATION
    if opt.max_duration and opt.max_duration <= max_duration then
        max_duration = opt.max_duration
    end

    local params = {
        filename = opt.filename or 'noname',
        format = opt.format or 'amr',
        max_filesize = max_filesize,
        max_duration = max_duration,
        sampling_rate = opt.sampling_rate or Media.DEFAULT_SAMPLING_RATE,
        encoding_bit_rate = opt.encoding_bit_rate or Media.DEFAULT_ENCODING_BIT_RATE,
        channel = opt.channel or Media.DEFAULT_CHANNEL,
    }

    if params.format == 'amr' then
        if params.sampling_rate ~= 8000 then
            error('when format is amr, sampling_rate should be 8000');
            return
        end
        if params.encoding_bit_rate ~= 16 then
            error('when format is amr, encoding_bit_rate should be 16');
            return
        end
        if params.channel ~= 1 then
            error('when format is amr, channel should be 1');
            return
        end
    end

    local cb_wrap = function(info, body)
        info = JSON:decode(info)
        cb(info, body);
    end

    local optStr = (opt and JSON:encode(params))
    M.async_call('media_start_record', cb_wrap, optStr)
end

function Media.stop_record(opt, cb)
    local optStr = (opt and JSON:encode(opt)) or '{}'
    local cb_wrap = function(info, body)
        info = JSON:decode(info)
        info.bytes = body
        cb(info);
    end
    M.async_call('media_stop_record', cb_wrap, optStr)
end

function Media.start_play(opt, cb)
    opt = opt or {}

    local format = 'amr'
    if opt.format ~= 'auto' then
        format = opt.format
    end

    local params = {
        filename = opt.filename or 'noname',
        format = format,
    }

    local cb_wrap = function(info, body)
        info = JSON:decode(info)
        cb(info, body);
    end

    local optStr = (opt and JSON:encode(params))
    M.async_call('media_start_play', cb_wrap, optStr);
end

function Media.stop_play(opt, cb)
    local optStr = (opt and JSON:encode(opt)) or '{}'
    local cb_wrap = function(info, body)
        info = JSON:decode(info)
        cb(info, body);
    end
    M.async_call('media_stop_play', cb_wrap, optStr);
end

function Media.delete(opt, cb)
    local optStr = (opt and JSON:encode(opt)) or '{}'
    local cb_wrap = function(info, body)
        info = JSON:decode(info)
        cb(info, body);
    end
    M.async_call('media_delete', cb_wrap, optStr);
end

function Media.get_record_dir()
    return _ejoysdk.media_record_dir()
end

return M
