local E = require "ejoysdk_lua.ejoysdk"
local ET = require "ejoysdk_lua.ejoysdk_topic"
local EU = require "ejoysdk_lua.ejoysdk_utils"
local Vendor = require "ejoysdk_lua.vendors.vendor"

local CHANNEL = "EJOY"
local EJOY = Vendor:Inherit(CHANNEL)

EJOY.EC_OAUTH_FAILED = -2
EJOY.EC_USER_CANCEL = -3
EJOY.EC_OAUTH_ERROR_CALLBACK_URL = -4
EJOY.EC_NONCE_ERROR = -5

local callback_info = {
    nonce = nil,
    type = 'browser',  -- 现在这个 type 也被用来作为 webview开关的标识，需要等接口
}

local ejoy_token = nil

local function ejoy_oauth_url()
    local url_base = E.CONFIG.get_config('id')
    return url_base .. '/o/oauth2/v1/auth'
end

local cb_handler = function(succ, token, new_nonce, _new_state)
    local nonce = callback_info.nonce

    -- webview 下nonce 不可能是空
    if callback_info.type ~= 'browser' then
        nonce = nonce or true
    end

    -- browser 下，可能游戏退出了，导致nonce变成了初始化的nil
    if not succ then
        ejoy_token = nil
        local error_info = token
        EJOY.opt.auth_listener(false, error_info)
    elseif (nonce and new_nonce ~= nonce) then
        ejoy_token = nil
        EJOY.opt.auth_listener(false, {code=EJOY.EC_NONCE_ERROR, msg="nonce error"})
    else
        ejoy_token = token
        EJOY.opt.auth_listener(succ, {
            platform = CHANNEL,
            ptoken = token,
            pid = nil,
            guest = false,
            with = nil,
            with_account = nil,
        }, {})
    end
end

local function ejoy_oauth_ticket_url()
    local url_base = E.CONFIG.get_config('id')
    return url_base .. '/app/api/v1/gen_ticket'
end

local function ejoy_catch_token_url(env)
    local os = E.CONFIG.get_config('os')
    local url_base = E.CONFIG.get_config('launcher')
    return url_base .. '/oauth/ejoy/' .. os .. '/' .. env
end

local function on_url(url)
    local ret = E.HTTP.parse(url)
    if ret.host == 'sdk.ejoy.com' and ret.query then
        local query = ret.query
        if query._platform == 'ejoy' and query.access_token and query.nonce and query.state then
            cb_handler(true, query.access_token, query.nonce, query.state)
        else
            if query.error then
                -- log error msg
                E.log(query)
                cb_handler(false, {code=EJOY.EC_OAUTH_FAILED, msg="ejoy id server: " .. tostring(query.error)})
            else
                cb_handler(false, {code=EJOY.EC_OAUTH_ERROR_CALLBACK_URL, msg="ouath error callback url"})
            end
        end
    end
end

local function on_login_done(value)
    callback_info.type = 'browser'
    on_url(value.args.uri);
end

local function on_webview_close(_value)
    if callback_info.type ~= 'browser' then
        callback_info.type = 'browser'
        cb_handler(false, {code=EJOY.EC_USER_CANCEL, msg="user canceled"})
    end
end

local EJOY_DEFAULT_OPT = {
    type = 'webview',
    hide_email_login = false,
}
EJOY_DEFAULT_OPT.__index = EJOY_DEFAULT_OPT

local function init_opt(opt)
    assert(type(opt) == 'table', "opt should be table")
    return setmetatable(opt, EJOY_DEFAULT_OPT)
end

function EJOY.init(opt)
    EJOY.opt = init_opt(opt)

    ET.subscribe('urlopen', on_url)
    ET.subscribe('logindone', on_login_done)
    ET.subscribe('webview_close', on_webview_close)
end

function EJOY.has_token()
    return ejoy_token
end

function EJOY.open_webview(webview_url)
    local secret = _ejoysdk_crypt.randomkey()
    local query = {
        secret = _ejoysdk_crypt.base64encode(_ejoysdk_crypt.dhexchange(secret)),
    }

    local ticket_url = E.HTTP.url_query(ejoy_oauth_ticket_url(), query)

    E.HTTP.get(ticket_url, {acceptable = E.HTTP.CT_JSON}, function(resp)
        if resp.status ~= 200 or resp.body.code ~= 0 then
            return cb_handler(false, {code=EJOY.EC_OAUTH_FAILED, msg="oauth failed"})
        end

        local server_secret = _ejoysdk_crypt.base64decode(resp.body.secret)
        local ticket_encrypted = _ejoysdk_crypt.base64decode(resp.body.ticket)

        local real_secret = _ejoysdk_crypt.dhsecret(server_secret, secret)
        local c = EU.XORCipher.new(real_secret)
        local ticket = c:decrypt(ticket_encrypted)

        local opt = {
            compactMode= true,
            closeEventData= 'oauth',
        }
        E.WebView.open(webview_url, {
            [".ejoy.com"] = {
                startupData= {
                    ejoy_ticket=_ejoysdk_crypt.base64encode(ticket),
                    hide_email_login= EJOY.opt.hide_email_login,
                },
                transparent= true,
            }
        }, opt)
    end)
end

function EJOY.check_token(outsource, info)
    cb_handler(true, outsource, info)
end

function EJOY.simple_token()
    return true
end

function EJOY.merge_info(info, pinfo)
    return EJOY.merge_helper(info, pinfo)
end

function EJOY.login(_outsource, _info)
    local viewer_type = EJOY.opt.type

    local nonce = _ejoysdk_crypt.base64encode(_ejoysdk_crypt.randomkey())
    local query = {
        response_type = 'token',
        client_id = E.CONFIG.get_config('product'),
        nonce = nonce,
        redirect_uri = ejoy_catch_token_url(viewer_type),
        scope = 'profile',
        state = 'login',
        login_hint = "CLEAN",
        v = 2,
    }

    callback_info = {
        nonce = nonce,
        type = viewer_type,
    }

    local url = E.HTTP.url_query(ejoy_oauth_url(), query)
    if viewer_type == 'browser' then
        E.Sysinfo.open_url(url)
    else
        EJOY.open_webview(url)
    end
end

function EJOY.logout(cb)
    cb()
end

EJOY:is_implemented({"ACCOUNT"})

return EJOY
