local E = require "ejoysdk_lua.ejoysdk"
local JSON = require "ejoysdk_lua.ejoysdk_json"

local UNISDK_EVENT = "UNISDK_EVENT"

local ACT_UNISDK_ASYNC_CALL = "UNISDK_ASYNC_CALL"

local IVK_UNISDK_LOGIN = "UNISDK_LOGIN"
local IVK_UNISDK_LOGOUT = "UNISDK_LOGOUT"
local IVK_UNISDK_PAY = "UNISDK_PAY"
local IVK_UNISDK_CAST = "UNISDK_CAST"

local CT_UNISDK_GET_ABILITY = "UNISDK_GET_ABILITY"
local CT_UNISDK_SYNC_CALL = "UNISDK_SYNC_CALL"

local EVT_LOGIN = 1
local EVT_LOGOUT = 2
local EVT_PAY = 3
local EVT_EVENT = 4

local M = {}

local nop = function()
    _ejoysdk.log("unisdk nop function called")
end

local cb_mt = {
    __index = function()
        return nop
    end
}

local login_info = {}
local login_cb_tbl = setmetatable({}, cb_mt)
local logout_cb_tbl = setmetatable({}, cb_mt)
local pay_cb_tbl = setmetatable({}, cb_mt)
local event_cb_tbl = setmetatable({}, cb_mt)

local function on_login(succ, chn, info, ext_params)
    if succ then
        login_info[chn] = {info = info, ext = ext_params}
    end
    login_cb_tbl[chn](succ, info, ext_params)
end

local function on_logout(chn, ext_params)
    login_info[chn] = nil
    logout_cb_tbl[chn](ext_params)
end

local function on_pay_result(succ, chn, order_id, ext)
    pay_cb_tbl[chn](succ, order_id, ext)
end

local function on_event(chn, type_, body)
    event_cb_tbl[chn](type_, body)
end

local API = {}
do
    local unisdk_config = "unisdk/sdkconfig.json"

    if E.CONFIG.get_config("os") == "android" then
        function API.login(chn, params)
            E.invoke(IVK_UNISDK_LOGIN, {channel = chn, params = params})
        end

        function API.logout(chn, params)
            E.invoke(IVK_UNISDK_LOGOUT, {channel = chn, params = params})
        end

        function API.pay(chn, order_id, params)
            E.invoke(
                IVK_UNISDK_PAY,
                {
                    channel = chn,
                    orderId = order_id,
                    params = params
                }
            )
        end

        function API.get_sdk(ability)
            return E.sync_call(CT_UNISDK_GET_ABILITY, {ability = ability})
        end

        function API.async_call(channel, type_, params, chunk, cb)
            params["channel"] = channel
            params["type"] = type_
            E.async_call(
                ACT_UNISDK_ASYNC_CALL,
                params,
                chunk,
                function(body, resp_chunk)
                    if body.code == 0 then
                        cb(true, body.body, resp_chunk)
                    else
                        cb(false, body.code, body.body, resp_chunk)
                    end
                end
            )
        end

        function API.sync_call(channel, type_, params, chunk)
            params["channel"] = channel
            params["type"] = type_
            return E.sync_call(CT_UNISDK_SYNC_CALL, params, chunk)
        end

        function API.cast(channel, type_, params, chunk)
            params["channel"] = channel
            params["type"] = type_
            E.invoke(IVK_UNISDK_CAST, params, chunk)
        end

        local event_dispatch = {
            [EVT_LOGIN] = function(value)
                E.log(value)
                if value.succ then
                    local info = {
                        channel = value.channel,
                        channel_product_code = value.channel_product_code,
                        token = value.token,
                        pid = value.user_id
                    }
                    on_login(true, value.channel, info, value.ext_params)
                else
                    local info = {
                        code = value.code,
                        msg = value.msg
                    }
                    on_login(false, value.channel, info, value.ext_params)
                end
            end,
            [EVT_LOGOUT] = function(value)
                on_logout(value.channel, value.ext_params)
            end,
            [EVT_PAY] = function(value)
                if value.succ then
                    on_pay_result(true, value.channel, value.order_id, value.ext)
                else
                    local ext = value.ext
                    ext.error_code = value.error_code
                    ext.error_message = value.error_message
                    on_pay_result(false, value.channel, value.order_id, ext)
                end
            end,
            [EVT_EVENT] = function(value)
                local chn = value.channel
                local type_ = value.type
                local body = value.body
                on_event(chn, type_, body)
            end
        }

        _ejoysdk.register_cb(
            UNISDK_EVENT,
            function(cbid, js_str, chunk)
                local handler = event_dispatch[cbid]
                if handler then
                    local value = JSON:decode(js_str)
                    handler(value, chunk)
                end
            end
        )
    else
        unisdk_config = "ejoysdk/sdkconfig.json"

        -- iOS
        API.login = function(chn, params)
            _ejoysdk.unisdk_login(chn, JSON:encode(params))
        end

        API.logout = function(chn, params)
            _ejoysdk.unisdk_logout(chn, JSON:encode(params))
        end

        API.pay = function(chn, order_id, params)
            _ejoysdk.unisdk_pay(chn, order_id, JSON:encode(params))
        end

        function API.get_sdk(ability)
            local sdkconfig = E.CONFIG.get_config("unisdk_meta")
            local ret = {sdks={}}
            for _, i in ipairs(sdkconfig.sdks) do
                for _, j in ipairs(i.ability) do
                    if j == ability then
                        table.insert(ret.sdks, i.channel)
                    end
                end
            end
            return ret
        end

        function API.async_call(channel, type_, params, chunk, cb)
            chunk = chunk or ""
            E.async_call(
                "unisdk_async_call",
                function(body, resp_chunk)
                    _ejoysdk.log("unisdk_async_callback: " .. body)
                    body = JSON:decode(body)
                    if body.code == 0 then
                        cb(true, body.body and JSON:decode(body.body), resp_chunk)
                    else
                        cb(false, body.code, body.body, resp_chunk)
                    end
                end,
                channel,
                type_,
                JSON:encode(params),
                chunk
            )
        end

        API.sync_call = function(chn, type_, params, chunk)
            chunk = chunk or ""
            local ret = _ejoysdk.unisdk_sync_call(chn, type_, JSON:encode(params), chunk)
            return JSON:decode(ret)
        end

        API.cast = function(chn, type_, params, chunk)
            chunk = chunk or ""
            _ejoysdk.unisdk_cast(chn, type_, JSON:encode(params), chunk)
        end

        local event_dispatch = {
            [EVT_LOGIN] = function(chn, status_code, body)
                local value = JSON:decode(body)
                if status_code == 0 then
                    local info = {
                        channel = chn,
                        channel_product_code = value.channel_product_code,
                        token = value.token,
                        pid = value.user_id
                    }
                    on_login(true, chn, info, value.ext_params)
                else
                    local info = {
                        code = value.code,
                        msg = value.msg
                    }
                    on_login(false, chn, info, value.ext_params)
                end
            end,
            [EVT_LOGOUT] = function(chn, _status_code, body)
                local value = JSON:decode(body)
                on_logout(chn, value.ext_params)
            end,
            [EVT_PAY] = function(chn, status_code, body)
                local value = JSON:decode(body)
                if status_code == 0 then
                    on_pay_result(true, chn, "", value)
                else
                    local ext = value
                    if status_code == 1 then
                        ext.error_code = "PAY_FAILED"
                    elseif status_code == 2 then
                        ext.error_code = "PAY_CANEL"
                    end
                    on_pay_result(false, chn, "", ext)
                end
            end,
            [EVT_EVENT] = function(chn, type_, body)
                on_event(chn, type_, JSON:decode(body).body)
            end
        }

        _ejoysdk.register_cb(
            UNISDK_EVENT,
            function(cbid, ...)
                local handler = event_dispatch[cbid]
                if handler then
                    handler(...)
                end
            end
        )
    end

    local data = _ejoysdk.lread(unisdk_config) or ""
    _ejoysdk.log("load config: \n" .. data)

    if data then
        E.CONFIG.set_config("unisdk_meta", JSON:decode(data))
    end
end

function M.login(chn, params)
    params = params or {}
    API.login(chn, params)
end

function M.logout(chn, params)
    params = params or {}
    login_info[chn] = nil
    API.logout(chn, params)
end

function M.get_info(chn)
    return login_info[chn]
end

function M.register_login_listener(channel, cb)
    login_cb_tbl[channel] = cb
end

function M.register_pay_listener(channel, cb)
    pay_cb_tbl[channel] = cb
end

function M.register_event_cb(channel, cb)
    event_cb_tbl[channel] = cb
end

function M.pay(chn, order_id, params)
    API.pay(chn, order_id, params)
end

function M.get_sdk(ability)
    return API.get_sdk(ability)
end

M.async_call = API.async_call
M.sync_call = API.sync_call
M.cast = API.cast

return M
