local E = require "ejoysdk_lua.ejoysdk"
local HTTP = E.HTTP
local ET = require "ejoysdk_lua.ejoysdk_topic"
local EG = require "ejoysdk_lua.ejoysdk_gangplank"
local UNI = require "ejoysdk_lua.vendors.unisdk"
local Vendor = require "ejoysdk_lua.vendors.vendor"

local CHANNEL = "WEIXIN"
local WEIXIN_SCOPE = "snsapi_userinfo"
-- 30 天 refresh_token 过期， 提前两天触发刷新
local WEIXIN_REFRASH_TOKEN_EXPIRES = (30 - 2) * 24 * 3600
local WEIXIN_REFRASH_TOKEN_URL = "https://api.weixin.qq.com/sns/oauth2/refresh_token"

local auth_listener = nil
local pay_listener = nil
local inited = false
local product_infos = {}

local WeiXinKS = E.LazyKeyStore:New('WEIXIN_TOKEN', false, true)
ET.subscribe("app_on_stop", function()
    if WeiXinKS:get() then
        WeiXinKS:save()
    end
end)

local M = Vendor:Inherit(CHANNEL)

local login_handler = function()
    if inited then
        return
    end

    _ejoysdk.log("get weixin info")

    EG.product_infos_base(CHANNEL, function(succ, infos)
        if succ then
            product_infos = infos
            E.log(product_infos)
            inited = true
            ET.publish('purchase_inited', CHANNEL)
        else
            _ejoysdk.log("log alipay product info fail")
        end
    end)
end

function M.init(opt)
    pay_listener = opt.pay_listener
    auth_listener = opt.auth_listener

    if inited then
        return
    end

    ET.subscribe('login', login_handler)
    _ejoysdk.log("init WEIXIN")

    UNI.register_pay_listener(CHANNEL, function(succ, order_id, _ext_params)
        if succ then
            pay_listener(succ, order_id)
        else
            pay_listener(succ)
        end
    end)
    UNI.register_login_listener(CHANNEL, function(succ, info, ext_params)
        if succ then
            if WeiXinKS:get().nonce == ext_params.nonce then
                WeiXinKS:delete()
                M.exchange_token(info.token, info, ext_params)
            end
        else
            auth_listener(succ)
        end
    end)
end

function M.login()
    local url = EG.gangplank_url("/notify/weixin/nonce")
    HTTP.get(url, {}, function(resp)
        if resp.status == 200 then
            local nonce = resp.body
            WeiXinKS:set({nonce = nonce})
            UNI.login(CHANNEL, {nonce = nonce, scope = WEIXIN_SCOPE})
        else
            auth_listener(false)
        end
    end)
end

function M.logout(cb)
    WeiXinKS:delete()
    cb()
end

function M.merge_info(info, pinfo)
    return M.merge_helper(info, pinfo)
end

function M.simple_token()
    return false
end

function M.check_token(outsource, info)
    local weixin_info = WeiXinKS:get()
    if weixin_info.token == outsource.ptoken then
        local now = os.time()
        if now > weixin_info.expires_in + weixin_info.now or
            now > WEIXIN_REFRASH_TOKEN_EXPIRES + weixin_info.now then
            M.refresh_token(weixin_info, outsource, info)
        else
            auth_listener(true, outsource, info)
        end
    else
        M.login()
    end
end

function M.can_pay()
    return inited
end

function M.pay(product_id, _count, order_id, body)
    assert(product_infos[product_id], "product_id " .. tostring(product_id) .. " not found")

    if body.cb.code ~= 'SUCCESS' then
        pay_listener(false)
        return
    end

    local order_info = {
        product_id = product_id,
        wxdata = body.cb.wx,
    }
    UNI.pay(CHANNEL, order_id, order_info)
end

function M.product_list()
    return product_infos
end

-- 内部函数
function M.exchange_token(code, user_info, ext_params)
    local url = EG.gangplank_url("/notify/weixin/access_token")
    local params = {
        code = code,
        nonce = ext_params.nonce,
    }
    HTTP.post(url, {acceptable=HTTP.CT_JSON}, HTTP.CT_URLENCODED , params, function(resp)
        if resp.status == 200 and resp.body.status == 200 then
            local body = resp.body
            WeiXinKS:set({
                token = body.access_token,
                expires_in = body.expires_in,
                refresh_token = body.refresh_token,
                pid = body.openid,
                unionid = body.unionid,
                lang = ext_params.lang,
                country = ext_params.country,

                now = os.time(),
            })
            local outsource = {
                platform = user_info.channel,
                ptoken = body.access_token,
                pid = body.openid,
                guest = false,
                with = nil,
                with_account = nil,
                appname = user_info.channel_product_code,
            }
            local info = {
                country = ext_params.country,
                lang = ext_params.lang,
            }

            auth_listener(true, outsource, info)
        else
            auth_listener(false)
        end
    end)
end

function M.refresh_token(weixin_info, outsource, info)
    local params = {
        appid = outsource.appname,
        grant_type = 'refresh_token',
        refresh_token = weixin_info.refresh_token,
    }

    HTTP.post(
        WEIXIN_REFRASH_TOKEN_URL, {acceptable=HTTP.CT_JSON},
        HTTP.CT_URLENCODED, params, function(resp)
            if resp.status == 200 and resp.body.errcode == nil then
                local body = resp.body
                weixin_info.token = body.access_token
                weixin_info.expires_in = body.expires_in
                weixin_info.refresh_token = body.refresh_token
                weixin_info.now = os.time()

                WeiXinKS:set(weixin_info)

                outsource.ptoken = body.access_token
                auth_listener(true, outsource, info)
            else
                WeiXinKS:delete()
                M.login()
            end
        end)
end

M:is_implemented({"ACCOUNT", "PAY"})

return M
